###############################################################################
###############################################################################
#                               Getleft
###############################################################################
###############################################################################
# copyright (C) 1999-2000 Andres Garcia Garcia - fandom@retemail.es
# Distributed under the terms of the GPL v2
###############################################################################
###############################################################################
# Version 0.10.1
###############################################################################
###############################################################################
# This program recursively downloads Web sites, as it goes, it changes absolute
# paths in the links to relative ones, and links to active content pages get
# changed to the resulting pages.
###############################################################################
#
# The following lines look for the most recent wish version in your path,
# keep in mind this program only works with version 8.1 or newer
#
# \
if which wish8.3&>/dev/null; then
# \
    exec wish8.3 -f $0
# \
elif which wish8.2&>/dev/null; then
# \
    exec wish8.2 -f $0
# \
elif which wish8.1&>/dev/null; then
# \
    exec wish8.1 -f $0
# \
else
# \
    exec wish -f $0
# \
fi

###############################################################################
# CheckPrograms
#    Checks whether the programs Geleft relies on: curl, cat, touch and sed are
#    within reach.
#
# Side effects:
#    If one of the programs is not in the path, the procedure will exit the
#    program after issuing a error message.
###############################################################################
proc CheckPrograms {} {
    global tcl_version errorCode

    if {$tcl_version<8.1} {
        tk_messageBox -icon error -type ok -message "This program only works with\n\
                Tcl/Tk v 8.1 or newer, yours is:\n\t $tcl_version" -title Error
        exit
    }
    set dependencies 0
    catch {exec curl}
    if {[lindex $errorCode 1]=="ENOENT"} {
        tk_messageBox -type ok -icon error -title "No curl" \
            -message "Getleft depends on program\n\tcurl\nPlease check the docs"
        exit
    }
    return
}

###############################################################################
# InitializeDownload
#    Takes care of downloading the Url entered by the user starting the
#    download.
#
# Parameter:
#    url: The requested Url.
#
# Returns:
#    The requested url, completed, if necessary with protocol or changed
#    if there was a redirection.
#    An empty string in case of error.
###############################################################################
proc InitializeDownload {url} {
    global directories siteUrl options
    global labelTitles labelMessages

    if {![regexp -nocase {([^:/])(/)} $url]} {
        set url $url/
    }

    set siteUrl(url) $url
    set parsedUrl [HtmlParser::ParseUrl $url]
    if {$parsedUrl!=1} {
        set siteUrl(prot) [lindex $parsedUrl 0]
        set siteUrl(www)  [lindex $parsedUrl 1]
        set siteUrl(dir)  [lindex $parsedUrl 2]
        set fileName      [lindex $parsedUrl 3]
        if {$fileName==""} {
            set isHtml 1
        } else {
            set isHtml [regexp -nocase {\.html?$} $fileName]
        }
        if {$siteUrl(prot)==""} {
            set siteUrl(prot) http
            set url http://$url
        } else {
            set siteUrl(prot) [string tolower $siteUrl(prot)]
        }
    } else {
        tk_messageBox -type ok -title $labelTitles(initError) -icon error \
            -message $labelMessages(formatError)
        return
    }

    set directories(local)  $directories(base)
    set siteUrl(base)       $siteUrl(dir)

    set result [Descargar::Download $url "-" $isHtml]

    if {$result==2} {
        tk_messageBox -icon error -type ok -title $labelTitles(error) \
                 -message "$Ventana::Rizo::errorMessage"
        return
    }

    if {($result==1)&&($isHtml==0)} {
        set localFileName [UrlToFile $url]
        if {![file exists $localFileName.html]} {
            file rename -force $localFileName $localFileName.html
            Commands::Touch $localFileName
        }
        return $url
    }

    if {($result==3)||($result==4)} {
        set parsedUrl [HtmlParser::ParseUrl $Descargar::relocated(url)]
        if {$parsedUrl==1} {
            return
        }
        set siteUrl(prot) [lindex $parsedUrl 0]
        set siteUrl(www)  [lindex $parsedUrl 1]
        set siteUrl(dir)  [lindex $parsedUrl 2]
        set fileName      [lindex $parsedUrl 3]

        set siteUrl(base) $siteUrl(dir)
        set url $Descargar::relocated(url)
    }
    return $url
}

###############################################################################
# RelativePath
#    The function returns the relative path from the referer page to the linked
#    page.
#
# Parameter:
#    urlRef. The referer page.
#    urlNew: The url whose link we are calculating.
#
# Returns:
#    The link for the changed page.
###############################################################################
proc RelativePath {urlRef urlNew} {
    global directories siteUrl

    set fileRef [UrlToFile $urlRef]
    set fileNew [UrlToFile $urlNew]
#    if {![regexp "://$siteUrl(www)/" $urlNew]} {
#        set external 1
#    } else {
#        set external 0
#    }

    regexp -nocase "(?:^$directories(base)/)(.*)" $fileRef nada fileRef
    regexp -nocase "(?:^$directories(base)/)(.*)" $fileNew nada fileNew

    set listDirRef [file dirname [split $fileRef /]]
    regsub -all {\+} $fileNew {\\+} fileNew
    regsub -all {\+} $fileRef {\\+} fileRef
    foreach dir $listDirRef {
        if {[regexp "(?:^$dir/)(.*)" $fileNew nada fileNew]} {
             regexp "(?:^$dir/)(.*)" $fileRef nada fileRef
        } else {
            break
        }
    }
    set jumps [regsub -all {/} $fileRef {} nada]
#    if {$external==1} {incr jumps}
    for {set i 0} {$i<$jumps} {incr i} {
        set fileNew ../$fileNew
    }
    regsub -all {\\\+} $fileNew {+} fileNew

    return $fileNew
}

###############################################################################
# RelativePath
#    The function returns the relative path from the current directory to the
#    root directory of the download.
#
# Parameter:
#    The directory of the file which is going to be processed for links.
#
# Returns:
#    The relative path.
###############################################################################
proc RelativePathOld {path_absoluto} {
    global directories

    set patron (?:$directories(base)/)(.*)
    regexp $patron $path_absoluto nada path
    if {![info exists path]} return

    set cuenta [regsub -all {/} $path {} ignorar]

    for {set i 0;set rel ""} {$i<$cuenta} {incr i} {
        append rel ../
    }
    set rel [string trimright $rel "/"]

    return $rel
}

###############################################################################
# Preprocessing
#    Reads the Web page passed as a parameter and proccess it to extract
#    all links to local files and images it has
#
# Parameters:
#    url: url of the page where are going to process for links.
#    level: Level of recursion in which the file is processed.
###############################################################################
proc Preprocessing {url level} {
    global directories getleftState charSets filesDone

    if [string match $url ""] return

    set file [UrlToFile $url]
    set filesDone($file) $level

    if {$::DEBUG==1} {
#        puts "Fichero a preprocesar: $file"
    }

    set directories(local) [file dirname $file]

    if {$getleftState(filesChosen)==1} {
        Ventana::ShowWindow  process
        Ventana::FileStrings $url
    }

    if {[file exists $file.orig]} {
        set fileName $file.orig
    } else {
        set fileName $file
    }

    set returnCode [HtmlParser::Parsing $fileName $url]
    if {$returnCode==1} {
        if {$::DEBUG==1} {
            tk_messageBox -type ok -icon error -type ok \
                    -message "No se pudo procesar el fichero: $fileName"
        }
        return
    }

    if {$getleftState(filesChosen)==0} {
        set encoding ""
        catch {set encoding $charSets($Ventana::Rizo::meta(charSet))}
        if {$HtmlParser::pageEncoding!=""} {
            if [catch {set $charSets($HtmlParser::pageEncoding)} encoding] {
                set encoding ""
            }
        }
        if {$encoding!=""} {
            HtmlParser::ChangeEncoding $encoding
        }
        HtmlParser::TidyDescription
    }
    if {[file exists $file.orig]} {
        file rename -force $file.orig $file
    }

    HtmlParser::FilterLinks $url
    update
    ChangePage $url

    return
}

###############################################################################
# TidyNames
#    Removes from the name and path of files things like '?' '~' '+' '-'
#
# Returns
#    The filename without those characters.
###############################################################################
proc TidyNames {nombre} {

    regsub -all {~}  $nombre {} nombre
    regsub -all {\*} $nombre {} nombre
    if {[regexp {(?:^.:)(.*)} $nombre nada filename]} {
        regsub -all {:}  $filename {} filename
	  set nombre $filename
    } else {
        regsub -all {:} $nombre {} nombre
    }
    if {[regexp {([^\?]+)(?:\?)(.*)} $nombre nada uno dos]} {
        regsub -all {\?} $dos {} dos
        regsub -all {\+} $dos {} dos
        regsub -all {/}  $dos {} dos
        regsub -all {\\} $dos {} dos
        set nombre $uno$dos
    }
    return $nombre
}

###############################################################################
# Chosen
#    Selects and deselects the chosen files.
#
# Parameters:
#    pattern: regular expresion that identifies the files to (de)select.
#    select: '1' to select, '0' to deselect.
#    map: 1 in case we are mapping the site.
###############################################################################
proc Chosen {pattern select {map 0} } {
    global siteMap

    for {set i 1} {$i<$HtmlParser::nLinks} {incr i} {
        if {$HtmlParser::links($i,shown)==0} continue
        if {$map==0} {
            if {[regexp -nocase $pattern $HtmlParser::links($i,file)]} {
                set HtmlParser::links($i,ok) $select
            }
        } else {
            if {[regexp -nocase $pattern $siteMap($i,1)]} {
                set HtmlParser::links($i,ok) $select
            }
        }
    }
    return
}

###############################################################################
# ChooseGroup
#    Activates or deactivates the links that are either before or after the
#    link where the user right-clicked.
#
# Parameters:
#    which: 'pre' the links before the click.
#           'fol' the links after.
#    action: '0' deactivates the links
#            '1' activates them.
#    path: Path of the text widget where the links are.
###############################################################################
proc ChooseGroup {which action path} {

    regexp {(.*)(\.)} [$path index insert] nada line
    if {$which=="pre"} {
        set next -1
    } else {
        set next  1
    }
    for {set i $line} {1} {incr i $next} {
        if {![winfo exists $path.enlaces$i]} {
	        if {($which=="pre")} {
		        if {$i==0} return else continue
			} else return
	    }
        if {$action==0} {
            $path.enlaces$i deselect
        } else {
            $path.enlaces$i select
        }
    }
    return
}

###############################################################################
# ChangeLinksToChoose
#    If the user changes the downloading options in the 'Choose Files' dialog
#    this procedure takes care of displaying the new links to choose from.
#
# Parameter:
#    referer: url of the page from wherer the links where taken.
#    text: The path of the text widget
#    map:  '1' if we are getting the site map
###############################################################################
proc ChangeLinksToChoose {referer text map} {
    global filesDone

    $text configure -cursor watch
    $text configure -state  normal
    $text delete 1.0 end

    set file [lindex [array get filesDone] 0]
    catch {unset filesDone}
    set refererFile [UrlToFile $referer]
    catch {file rename -force $refererFile.orig $refererFile}

    HtmlParser::FilterLinks $referer
    ChangePage $referer

    LinksToChoose $text $map

    $text configure -state  disabled
    $text configure -cursor left_ptr

    return
}

###############################################################################
# LinksToChoose
#    Inserts the links in the 'Choose files' directory.
#
# Parameter
#    text: path of the text widget.
#    map: '1' if we are getting the site map
###############################################################################
proc LinksToChoose {text map} {
    global siteIndex siteMap charSets

    set color [$text cget -background]

    if {$map==1} {
        set HtmlParser::nLinks $siteIndex
    }

    for {set i 1 ; set j 1 ; set pre ""} {$i<$HtmlParser::nLinks} {incr i} {
        set HtmlParser::links($i,shown) 0
    	if {$map==0} {
            if {$HtmlParser::links($i,ok)==0} continue
        } else {
            set HtmlParser::links($i,ok) 1
        }
        set HtmlParser::links($i,shown) 1
        set boton [checkbutton $text.enlaces$j -background $color\
                -variable HtmlParser::links($i,ok)]

        $text window create $i.0 -window $text.enlaces$j
        if {$map==1} {
            set description $siteMap($i,2)
            for {set k 0 ; set spc "   " ; set pre ""} {$k<$siteMap($i,0)} \
                    {incr k} {
                append pre $spc
            }
        } else {
            set description $::HtmlParser::links($i,descrip)
        }
        $text.enlaces$j configure -text $pre$description
        $text insert insert \n

        bind $boton <Button-3> "
            regexp {(?:enlaces)(.*)} %W nada line
            $text mark set insert \$line.0
            tk_popup .elegir.menu %X %Y
        "
        bind $boton <Button-4> "$text yview scroll -1 pages"
        bind $boton <Button-5> "$text yview scroll  1 pages"

        incr j
    }
    return
}

###############################################################################
# ChooseDownloads
#    Shows a dialog with all the links in the entered Url so you can choose
#    which ones will be followed and which will be ignored.
#
# Parameter
#    referer: url of the page from which the links where taken.
#    map: '1' if we are getting the site map, defaults to '0'
#
###############################################################################
proc ChooseDownloads {referer {map 0} } {
    global labelButtons labelTitles labelDialogs labelMenus
    global dirGetleft options optionsTmp done charSets getleftState


    catch {destroy .balloon}
    catch {wm withdraw $Ventana::window(top)}

    if {$HtmlParser::nLinks==1} "return 0"
    set elegir  [toplevel .elegir]
    set marco   [frame $elegir.marco]
    set checks  [frame $elegir.checks]
    set botones [frame $elegir.botones]

    wm title $elegir $labelTitles(choose)

    set texto [text $marco.texto -yscrollcommand [list $marco.yscroll set]    \
            -xscrollcommand [list $marco.xscroll set] -wrap none              \
            -cursor left_ptr]
    set aceptar [button $botones.aceptar   -textvariable labelButtons(ok)     \
            -width 9 -command "SetChosenLinks;set done 1"]
    set cancelar [button $botones.cancelar -textvariable labelButtons(cancel) \
            -width 9 -command "set done 0" ]

    set mas   [image create bitmap  -file [file join $dirGetleft(images) mas]]
    set menos [image create bitmap  -file [file join $dirGetleft(images) menos]]
    set lhtml [label  $checks.lhtml -text "  $labelDialogs(html)"]
    set htmla [button $checks.htmla -image $mas   -width 8 -height 8 \
            -command "Chosen {(html$)|(htm$)|(/$)} 1 $map"]
    set htmle [button $checks.htmle -image $menos -width 8 -height 8 \
            -command "Chosen {(html$)|(htm$)|(/$)} 0 $map"]
    set lima  [label $checks.lima   -text "  $labelDialogs(images)"]
    set imaa  [button $checks.imaa  -image $mas   -width 8 -height 8 \
            -command "Chosen {(gif$)|(jpg$)|(bmp$)|(xbm$)|(tiff$)|(png$)} 1 $map"]
    set imae  [button $checks.imae  -image $menos -width 8 -height 8 \
            -command "Chosen {(gif$)|(jpg$)|(bmp$)|(xbm$)|(tiff$)} 0 $map"]
    set lcom  [label $checks.lcom   -text "  $labelDialogs(archives)"]
    set coma  [button $checks.cona  -image $mas   -width 8 -height 8 \
            -command "Chosen {(tar$)|(gz$)|(z$)|(zip$)} 1 $map"]
    set come  [button $checks.cone  -image $menos -width 8 -height 8 \
            -command "Chosen {(tar$)|(gz$)|(z$)|(zip$)} 0 $map"]
    set ltodo [label $checks.ltodo  -text "  $labelDialogs(all)"]
    set todoa [button $checks.todoa -image $mas   -width 8 -height 8 \
            -command "Chosen {.} 1 $map"]
    set todoe [button $checks.todoe -image $menos -width 8 -height 8 \
            -command "Chosen {.} 0 $map"]

    LinksToChoose $texto $map

    set barray [scrollbar $marco.yscroll -orient vertical \
        -command [list $texto yview]]
    set barrax [scrollbar $marco.xscroll -orient horizontal  \
        -command [list $texto xview]]

    set menu [menu $elegir.menu -tearoff 0]
    $menu add cascade -label "$labelDialogs(html)"      -menu $menu.html
    $menu add separator
    $menu add cascade -label "$labelDialogs(images)"    -menu $menu.images
    $menu add separator
    $menu add cascade -label "$labelDialogs(archives)"  -menu $menu.archives
    $menu add separator
    $menu add cascade -label "$labelDialogs(all)" -menu $menu.all
    $menu add separator
    $menu add cascade -label "$labelDialogs(previous)"  -menu $menu.previous
    $menu add cascade -label "$labelDialogs(following)" -menu $menu.following
    $menu add separator
    $menu add cascade -label "$labelMenus(options)" -menu  $menu.options

    menu $menu.html      -tearoff 0
    menu $menu.images    -tearoff 0
    menu $menu.archives  -tearoff 0
    menu $menu.all       -tearoff 0
    menu $menu.previous  -tearoff 0
    menu $menu.following -tearoff 0
    menu $menu.options   -tearoff 0

    $menu.html      add command -label "$labelDialogs(activate)"   \
            -command "$htmla invoke"
    $menu.html      add command -label "$labelDialogs(deactivate)" \
            -command "$htmle invoke"
    $menu.images    add command -label "$labelDialogs(activate)"   \
            -command "$imaa  invoke"
    $menu.images    add command -label "$labelDialogs(deactivate)" \
            -command "$imae  invoke"
    $menu.archives  add command -label "$labelDialogs(activate)"   \
            -command "$coma  invoke"
    $menu.archives  add command -label "$labelDialogs(deactivate)" \
            -command "$come  invoke"
    $menu.all       add command -label "$labelDialogs(activate)"   \
            -command "$todoa invoke"
    $menu.all       add command -label "$labelDialogs(deactivate)" \
            -command "$todoe invoke"
    $menu.previous  add command -label "$labelDialogs(activate)"   \
            -command "ChooseGroup pre 1 $texto"
    $menu.previous  add command -label "$labelDialogs(deactivate)" \
            -command "ChooseGroup pre 0 $texto"
    $menu.following add command -label "$labelDialogs(activate)"   \
            -command "ChooseGroup fol 1 $texto"
    $menu.following add command -label "$labelDialogs(deactivate)" \
            -command "ChooseGroup fol 0 $texto"

    $menu.options add command                                   \
            -label $labelMenus(all)                             \
            -command "Opciones::ChooseOptions
                      ChangeLinksToChoose $referer $texto $map"
    $menu.options add cascade -menu $menu.options.directorios   \
            -label $labelMenus(uplinks)
    $menu.options add cascade -menu $menu.options.levels        \
            -label $labelMenus(levels)
    $menu.options add cascade -menu $menu.options.exLevels      \
            -label $labelMenus(external)
    $menu.options add cascade -menu $menu.options.filters       \
            -label $labelMenus(fileFilter)
    $menu.options add cascade -menu $menu.options.encodings     \
            -label $labelMenus(encodings)
    $menu.options add check   -variable options(update)         \
            -label $labelMenus(update)   -onvalue 1 -offvalue 0
    $menu.options add check   -variable options(cgi)            \
            -label $labelMenus(cgi)      -onvalue 1 -offvalue 0 \
            -command "ChangeLinksToChoose $referer $texto $map"

    set menuDir [menu $menu.options.directorios -tearoff 0     \
            -postcommand {set optionsTmp(dir) $options(dir)}]
    $menuDir add radio -variable options(dir)  -value 1        \
           -label $labelMenus(linksFollow) -command "
                if {\$optionsTmp(dir)!=1} {
                    ChangeLinksToChoose $referer $texto $map
                }
            "
    $menuDir add radio -variable options(dir)  -value 0 \
            -label $labelMenus(linksIgnore) -command "
                if {\$optionsTmp(dir)!=0} {
                    ChangeLinksToChoose $referer $texto $map
                }
            "
    set menuLev [menu $menu.options.levels -tearoff 0]
    $menuLev add radio -variable options(levels) -value -1 \
            -label $labelMenus(noLimit)
    for {set i 0} {$i<6} {incr i} {
        $menuLev add radio -label $i -variable options(levels) -value $i
    }
    set menuEx [menu $menu.options.exLevels -tearoff 0]
    $menuEx add radio -variable options(exLevels)    \
            -label $labelMenus(linksIgnore) -value 0 \
            -command "
                set options(external) 0
                ChangeLinksToChoose $referer $texto $map
            "
    for {set i 1} {$i<3} {incr i} {
        $menuEx add radio -label $i -variable options(exLevels) \
            -value $i -command "
                set options(external) 1
                ChangeLinksToChoose $referer $texto $map
            "
    }

    set menuFil [menu $menu.options.filters -tearoff 0]
    $menuFil add check -variable options(onlyHtml) -onvalue 1 -offvalue 0 \
            -label $labelMenus(onlyHtml)
    $menuFil add cascade -menu $menuFil.images \
            -label $labelMenus(images)
    $menuFil add command -label $labelMenus(chooseFilter) -command "
        Herramientas::FilterFiles
        update
        ChangeLinksToChoose $referer $texto $map
    "

    set menuFilImage [menu $menuFil.images -tearoff 0]
    $menuFilImage add radio -variable options(images) -value 0 \
	        -label $labelMenus(allImages)
    $menuFilImage add radio -variable options(images) -value 1 \
            -label $labelMenus(onlyThumb) -command "
                ChangeLinksToChoose $referer $texto $map
            "
    $menuFilImage add radio -variable options(images) -value 2 \
            -label $labelMenus(noThumb) -command "
                ChangeLinksToChoose $referer $texto $map
            "

    set menuEncodings [menu $menu.options.encodings -tearoff 0]
    set encodings {ascii korean iso-2022-kr euc-kr iso-2022-jp iso-8859-1   \
                   iso-8859-2 iso-8859-2 iso-8859-3 iso-8859-4 iso-8859-5   \
                   iso-8859-6 iso-8859-7 iso-8859-8 iso-8859-9 iso-ir-159   \
                   iso-ir-87 ibm437 ibm850 ibm852 ibm855 ibm857 ibm860     \
                   ibm861 ibm862 ibm863 ibm864 ibm865 ibm869 koi8-r}
    foreach code $encodings {
        $menuEncodings add radio -variable tmp -value $code \
                -label $code -command "
                    Preprocessing $referer 0
                    HtmlParser::ChangeEncoding \$charSets(\$tmp)
                    ChangeLinksToChoose $referer $texto $map
                "
    }

    bind $elegir   <KeyPress-Prior> "$texto yview scroll -1 pages"
    bind $elegir   <KeyPress-Next>  "$texto yview scroll  1 pages"
    bind $elegir   <Key-Up>         "$texto yview scroll -1 units"
    bind $elegir   <Key-Down>       "$texto yview scroll  1 units"
    bind $elegir   <MouseWheel> "
        if {%D<0} {
            $texto yview scroll  1 pages
        } else {
            $texto yview scroll -1 pages
        }
    "
    bind $elegir   <Escape>         "$cancelar invoke"
    bind $texto    <Button-3>       "
        tkTextButton1 %W %x %y
        tk_popup $menu %X %Y
    "
    bind $texto    <Button-1>       {
        %W tag remove sel 0.0 end
    }

    grid $marco -sticky news
    grid $texto $barray -sticky news
    grid $barrax -sticky ew
    grid $checks -sticky ew
    grid $lhtml $htmla $htmle $lima $imaa $imae $lcom $coma $come \
        $ltodo $todoa $todoe
    grid $botones -sticky ew -ipady 4
    grid $aceptar $cancelar  -padx 3

    grid rowconfigure    $elegir 0 -weight 1
    grid columnconfigure $elegir 0 -weight 1
    grid rowconfigure    $marco  0 -weight 1
    grid columnconfigure $marco  0 -weight 1

    $texto configure -state disabled

    wm protocol $elegir WM_DELETE_WINDOW "
            set done 0
            destroy $elegir"

    tkwait visibility $elegir

# Annoying bug in Window with grab and popup menus
#   grab $elegir
    tkwait variable done
#   grab release $elegir

    destroy $elegir

    return $done
}

###############################################################################
# UrlToFile {url} {
#    Given an Url this procedure will return the file in which it will be
#    saved.
#
# Parameters
#    url: the url to process.
#
# Returns:
#    The file in which it will be saved complete with full path.
###############################################################################
proc UrlToFile {url} {
    global directories siteUrl

    set parsedUrl [HtmlParser::ParseUrl $url]
    set prot      [lindex $parsedUrl 0]
    set domain    [lindex $parsedUrl 1]
    set dir       [lindex $parsedUrl 2]
    set file      [lindex $parsedUrl 3]

    if {$file==""} {
        if {$prot=="ftp"} {
            set file index.txt
        } else {
            set file index.html
        }
    }
    if {[string compare [string tolower $domain] \
            [string tolower $siteUrl(www)]]} {
        set fileName /${domain}$dir/$file
    } else {
        set fileName $dir/$file
    }
    set fileName [TidyNames $fileName]

    if [regexp {(?:^/)(.*)} $fileName nada fileName] {
        set directoryRoot $directories(base)
    } else {
        set directoryRoot $directories(local)
    }

    set fileName [file join $directoryRoot $fileName]
    if {[file exists $fileName]} {
        if {([file size $fileName]==0)&&([file exists $fileName.html])} {
            set fileName $fileName.html
        }
    }
    return $fileName
}

###############################################################################
# CheckDir
#    Creates, if needed, the directory in which the link will be saved.
#
# Parameter:
#    fileName: file in which the url is going to be downloaded.
#
# Returns:
#    1: In case of error
###############################################################################
proc CheckDir {fileName} {
    global labelTitles labelMessages

    set newDirectory [file dirname $fileName]
    if [file exists $newDirectory] {
        if {![file isdirectory $newDirectory]} {
            tk_messageBox -type ok -icon error -title $labelTitles(error) \
                    -message "File names are screwed: $fileName"
            return 1
        }
     } else {
        if [catch {file mkdir $newDirectory} error] {
            tk_messageBox -type ok -title $labelTitles(error) -icon error \
                    -message $error
            return 1
        }
     }
     return
}

###############################################################################
# ChangePage
#    Changes a html page, so that there is consistency with the local
#    directories. After this procedure is run through a page all it's links
#    should be between double qoutes (") and be relative to the the current
#    directory.
#
# Parameters
#    url: The url of the page about to be changed.
###############################################################################
proc ChangePage {url} {
    global siteUrl
    global directories

    if {$HtmlParser::baseTag!=""} {
        Commands::SedChangeEnter <$HtmlParser::baseTag> ""
    }

    for {set i 1} {$i<$HtmlParser::nLinks} {incr i} {
        if {$HtmlParser::links($i,ok)==0} continue
        set link    $HtmlParser::links($i,file)
        set newLink [RelativePath $url $HtmlParser::links($i,url)]

        Commands::SedChangeEnter "(=)(\'|\")($link)(\"|\')" "=\"$newLink\""
        Commands::SedChangeEnter "(=)($link)(\ |>)" "=\"$newLink\"\\3"
    }
    Commands::Sed [UrlToFile $url]

    return
}

###############################################################################
# SetChosenLinks
#    Puts into getleftState(chosenLinks) the links the user wants to download.
###############################################################################
proc SetChosenLinks {} {
    global options getleftState

    if {($options(map)==1)&&($getleftState(mapDone)==1)} {
    set getleftState(mapLinks) $getleftState(chosenLinks)
    }
    set getleftState(chosenLinks) ""
    for {set i 1} {$i<$HtmlParser::nLinks} {incr i} {
        lappend getleftState(chosenLinks) $HtmlParser::links($i,ok)
    }
    return
}

###############################################################################
# ReadSel
#    Reads the selection from the clipboard
#
# Returns
#    Whatever was in the clipboard
###############################################################################
proc ReadSel {} {
    if {[catch {selection get -selection CLIPBOARD} cosa]} {
        if {[catch {selection get} cosa]} return
    }
    return $cosa
}

###############################################################################
# TheEnd
#    After downloading is completed or stopped by the user, this procedure
#    takes care of telling the user all about it.
###############################################################################
proc TheEnd {} {
    global labelTitles labelMessages
    global getleftState dirGetleft
    global siteUrl options

    catch {destroy .balloon}
    if {[catch {wm withdraw $::Ventana::window(top)} cosa]} {
        if {$::DEBUG==1} {
            tk_messageBox -title Error -type ok -icon error \
                    -message "Error al ocultar .bajando:\n$cosa"
        }
    }

    set tmp [string tolower $siteUrl(www)]
    set index [lsearch -exact $options(resumeSites) $tmp]
    if {($options(stopFile)==1)||($options(stopPage)==1)} {
        set what [tk_messageBox -icon question -title $labelTitles(resumeSite)\
                -type yesno -message $labelMessages(resumeSite)]
        if {[string compare $what "no"]} {
            if {$index==-1} {
                lappend options(resumeSites) $tmp
                SetResumeSitesMenu
            }
            SaveDownloadData   $tmp
        } else {
            set options(resumeSites) [lreplace $options(resumeSites) \
                    $index $index]
            SetResumeSitesMenu
            catch {file delete [file join $dirGetleft(conf) $tmp.gdt]}
        }
    } else {
        tk_messageBox -icon info -title $labelTitles(theEnd) -type ok \
                -message $labelMessages(theEnd)
        if {$index!=-1} {
            set options(resumeSites) [lreplace $options(resumeSites) \
                    $index $index]
            SetResumeSitesMenu
            catch {file delete [file join $dirGetleft(conf) $tmp.gdt]}
        }
    }
    SaveConfig

    if {$getleftState(errorFound)==1} {
        ::Herramientas::ShowErrorLog
    }
    return
}

###############################################################################
# AutomaticDownload
#    Resumes a download where the user told Getleft to leave it.
#
# Parameter:
#    Domain: The domain of the site to resume downloading.
###############################################################################
proc AutomaticDownload {domain} {
    global options siteUrl siteIndex siteMap urlsDownloaded
    global directories dirGetleft getleftState
    global labelMessages labelTitles

    set optionsFile [file join $dirGetleft(conf) $domain.gdt]
    if {[catch {source $optionsFile}]} {
        tk_messageBox -title $labelTitles(error) -icon error -type ok \
                -message $labelMessages(noResume)
        set index [lsearch $options(resumeSites) $domain]
        set options(resumeSites) [lreplace $options(resumeSites) \
                $index $index]
        SetResumeSitesMenu
        return
    }

    set options(stopPage)         0
    set options(stopFile)         0
    set options(pauseFile)        0
    set getleftState(filesChosen) 0
    set getleftState(errorFound)  0

    catch "unset ::filesDone"
    catch "unset ::urlsDownloaded"
    catch "unset ::Descargar::FilesNotFound"
    catch "unset ::Descargar::updated"

    set filesDoneFile [file join $dirGetleft(conf) $domain.gdt]
    if {[catch {source $filesDoneFile} result]} {
        tk_messageBox -type ok -icon error -title Error -message $result
    }

    set getleftState(downloading) 1
    set requestedUrl  [InitializeDownload $siteUrl(url)]
    if {$requestedUrl==""} {
        catch {destroy .balloon}
        catch {wm withdraw $Ventana::window(top)}
        set getleftState(downloading) 0

        return
    }
    set localFileName [UrlToFile $requestedUrl]

    if {($options(levels)==0)||(![regexp -nocase {\.html?$} $localFileName])} {
        TheEnd
        set getleftState(downloading) 0
        return $requestedUrl
    }
    Preprocessing $requestedUrl 0
    for {set i 1;set j 0} {$i<$HtmlParser::nLinks} {incr i;incr j} {
        if {($options(map)==0)||($getleftState(mapDone)==0)} {
            set HtmlParser::links($i,ok) [lindex $getleftState(chosenLinks) $j]
        } else {
            set HtmlParser::links($i,ok) [lindex $getleftState(mapLinks) $j]
        }
    }
    set getleftState(filesChosen) 1
    if {$options(map)==1} {
        set options(onlyHtml)     1
        set siteIndex             1
    }
    Descargar::ControlDownloading $requestedUrl 1
    if {$options(map)==0} {
        TheEnd
    } else {
        DownloadFromMap
        TheEnd
        unset siteMap
        unset HtmlParser::links
    }
    set options(stopFile)           0
    set options(stopPage)           0
    set options(pauseFile)          0
    set options(pausePage)          0
    set options(onlyHtml)           0
    set getleftState(filesChosen)   0
    set getleftState(downloading)   0

    return $requestedUrl
}

###############################################################################
# EnterUrlWindow
#    Creates a dialog box where you can write or paste the desired url.
#
# Returns:
#    The path of the entry in the combobox.
###############################################################################
proc EnterUrlWindow {} {
    global labelButtons labelTitles labelFrames
    global urlHistory

    set coord(x) [winfo rootx .]
    set coord(y) [winfo rooty .]

    set win [toplevel .enterUrlWindow]
    wm title $win $labelTitles(enterUrl)
    wm resizable $win 0 0
    wm geometry  $win +[expr $coord(x)+100]+[expr $coord(y)+15]

    set done  0

    set marcoEx [frame $win.marcoEx]
    set marco [fl::FrameLabel $marcoEx.marco -bd 2 -relief groove     \
            -textvariable labelFrames(webPage)]
    Ccombinado::ComboBox $marco.pagina $urlHistory -default [ReadSel] \
            -erasable 1

    set pagina $marco.pagina
    set botones  [frame $marcoEx.botones]
    set empezar  [button $win.empezar  -textvariable labelButtons(start)  \
            -width 8 -command {set done 1}]
    set cancelar [button $win.cancelar -textvariable labelButtons(cancel) \
            -width 8 -command {set done 0}]

    bind $pagina.entry <Return>   "$empezar invoke"
    bind $pagina.entry <KP_Enter> "$empezar invoke"
    bind $pagina.entry <Button-3> "
        $pagina.entry insert insert \[ReadSel\]
    "
    bind $win    <Escape>   "$cancelar invoke"

    pack $pagina   -padx 20 -pady 20
    pack $marcoEx  -ipadx 10 -ipady 5
    pack $marco    -side bottom
    pack $cancelar -side right -padx 9 -pady 5
    pack $empezar  -side right -pady 5
    pack $botones  -side right

    focus $pagina.entry

    return $pagina.entry
}

###############################################################################
# EnterUrl
#    Creates a dialog box where you can write or paste the desired url.
#
# Returns:
#    The url entered by the user.
###############################################################################
proc EnterUrl {} {
    global options done
    global directories dirGetleft getleftState
    global labelMessages labelTitles

    if [winfo exists .enterUrlWindow] {
        return
    }

    set options(stopPage)                 0
    set options(stopFile)                 0
    set options(pausePage)                0
    set getleftState(filesChosen)         0
    set getleftState(errorFound)          0
    set Ventana::Rizo::curlReport(stop)   0

    set comboEntry [EnterUrlWindow]

    tkwait variable done
    set requestedUrl [$comboEntry get]
    regsub -all {\n} $requestedUrl {} requestedUrl
    destroy .enterUrlWindow

    if {($done==0)||($requestedUrl=="")} {
        return
    }

    PutUrlInHistory $requestedUrl

    catch "unset ::filesDone"
    catch "unset ::urlsDownloaded"
    catch "unset ::Descargar::filesNotFound"
    catch "unset ::Commands::sedChanges"

    set directories(base) [Dialogos::SelectDirectory $dirGetleft(lastDir)]
    if {![string compare $directories(base) ""]} {
        return
    }
    set dirGetleft(lastDir) [file dirname $directories(base)]
    set getleftState(downloading) 1
    set requestedUrl  [InitializeDownload $requestedUrl]
    if {$requestedUrl==""} {
        catch {destroy .balloon}
        catch {wm withdraw $Ventana::window(top)}
        set getleftState(downloading) 0
        return
    }
    set localFileName [UrlToFile $requestedUrl]

    if {($options(levels)==0)||(![regexp -nocase {\.html?$} $localFileName])} {
        TheEnd
        set getleftState(downloading) 0
        return $requestedUrl
    }

    Preprocessing $requestedUrl 0

    if {$HtmlParser::nLinks==1} {
        tk_messageBox -icon info -title $labelTitles(noLinks) -type ok \
                -message $labelMessages(noLinks)
        set getleftState(downloading) 0
        catch {wm withdraw $::Ventana::window(top)}
        return
    }
    set chosen [ChooseDownloads $requestedUrl]
    if {$chosen==1} {
        Descargar::ControlDownloading $requestedUrl 1
        if {$options(map)==0} {
            TheEnd
        } else {
            return $requestedUrl
        }
        set options(stopFile)           0
        set options(stopPage)           0
        set options(pauseFile)          0
        set options(pausePage)          0
        set getleftState(filesChosen)   0
    }
    set getleftState(downloading) 0

    return $requestedUrl
}

###############################################################################
# DownloadFromMap
#    Download the links chosen in the map of the site
###############################################################################
proc DownloadFromMap {} {
    global siteIndex siteMap options getleftState

    set options(onlyHtml)  0
    set tmpLevels          $options(levels)
    set options(levels)    0

    if {$siteIndex==1} {
        return
    }

    if {$getleftState(mapDone)==0} {
        set getleftState(mapDone) 1
        set done [ChooseDownloads "" 1]
        if {$done==0} {
            set getleftState(downloading) 0
            return
        }
    } else {
        for {set i 1} {$i<$siteIndex} {incr i} {
            set HtmlParser::links($i,ok) [lindex $getleftState(chosenLinks) $i]
        }
    }

    for {set i 1} {$i<$siteIndex} {incr i} {
        if {$HtmlParser::links($i,ok)==1} {
            set enlace $siteMap($i,1)
            set filename [UrlToFile $enlace]
            if {[file exists $filename]} {
                continue
            } else {
                set isHtml [regexp -nocase {\.html?} $filename]
                Descargar::Download $enlace "-" $isHtml
            }
            if {$options(stopFile)==1} {
                 set options(levels) $tmpLevels
                 return
            }
            if {$options(pauseFile)==1} {
                Ventana::Pause file
            }
        }
    }
    set options(levels) $tmpLevels
    set options(map)    0

    return
}

###############################################################################
# GetSiteMap
#    Downloads the html pages in a site, to extract the site map from them
###############################################################################
proc GetSiteMap {} {
    global siteIndex siteMap options getleftState
    global siteUrl labelMessages labelTitles

    set options(map)          1
    set options(onlyHtml)     1
    set siteIndex             1
    set getleftState(mapDone) 0

    catch {unset siteMap}

    set requestedUrl [EnterUrl]
    if {$requestedUrl==""} return

    if {($options(stopFile)==1)||($options(stopPage)==1)} {
    } else {
        DownloadFromMap
    }
    TheEnd

    unset siteMap
    unset HtmlParser::links

    set options(stopFile)           0
    set options(stopPage)           0
    set options(pauseFile)          0
    set options(pausePage)          0
    set getleftState(filesChosen)   0
    set getleftState(downloading)   0
    set getleftState(mapDone)       0

    return
}

###############################################################################
# SaveDownloadData
#     Saves the data needed to automatically resume a download.
#
# Parameter
#     domain: The domain of the site.
###############################################################################
proc SaveDownloadData {domain} {
    global dirGetleft options siteUrl directories getleftState urlsDownloaded

    set fileName [file join $dirGetleft(conf) $domain.gdt]
    catch {file delete -force "$fileName"}
    set file [open "$fileName" w]

    puts $file "# Data to resume downloading $domain"
    puts $file "# This file is generated automatically, please do not edit"
    puts $file "\n\n"
    puts $file "set siteUrl(url)               \"$siteUrl(url)\""
    puts $file "set options(images)            $options(images)"
    puts $file "set options(cgi)               $options(cgi)"
    puts $file "set options(dir)               $options(dir)"
    puts $file "set options(external)          $options(external)"
    puts $file "set options(exLevels)          $options(exLevels)"
    puts $file "set options(filter)            \"$options(filter)\""
    puts $file "set options(levels)            $options(levels)"
    puts $file "set options(map)               $options(map)"
    puts $file "set options(onlyHtml)          $options(onlyHtml)"
    puts $file "set options(update)            $options(update)"
    puts $file "set directories(base)          \"$directories(base)\""
    puts $file "set getleftState(mapDone)      $getleftState(mapDone)"
    puts $file "set getleftState(chosenLinks)  \[list $getleftState(chosenLinks)\]"

    if {[info exists getleftState(mapLinks)]} {
        puts $file "set getleftState(mapLinks) \[list $getleftState(mapLinks)\]"
    }
    if {$options(map)==0} {
        puts $file "\n\n# Urls that have been already completed\n\n"
        foreach {url level} [array get urlsDownloaded] {
            puts $file "set urlsDownloaded($url) $level"
        }
    }
    if {$options(update)==1} {
        puts $file "\n\n# Urls that have been already updated\n\n"
        foreach {url nada} [array get Descargar::updated] {
            puts $file "set \"Descagar::updated($url)\" 1"
        }
    }
    puts $file "\n\n"
    close $file

    return
}

###############################################################################
# ReadConfig
#    Reads the configuration files
#
# Parameter:
#    Configuration file to read, if not present, the user configuration  file
#    will be read.
###############################################################################
proc ReadConfig { {confFile {} } } {
    global urlHistory urlIndex dirGetleft options
    global env tcl_platform


    if {$confFile!=""} {
        set ficherorc $confFile
    } else {
        if [regexp {Windows} $tcl_platform(os)] {
            set ficherorc [file join "$dirGetleft(main)" getleft.ini]
            set dirGetleft(lastDir) $dirGetleft(main)
        } else {
            set ficherorc [file join "$dirGetleft(conf)" getleft.conf]
            set dirGetleft(lastDir) $env(HOME)
            set temp [file join $env(HOME) .Getleft]
            if {[file exists $temp]} {
                set ficherorc $temp
                set update 1
            }
        }
    }

    set options(resumeSites) ""
    set urlHistory ""
    if [catch {open $ficherorc r} handle] {
        Commands::Touch $ficherorc
        set options(lang) en
        set urlIndex 0
        return 0
    }

    for {} {![eof $handle]} {} {
        set line [gets $handle]
        if [regexp {^#|^ |^$} $line] {
            continue
        }
        if [regexp {(?:\[)(.*)(?:\])} $line nada active] {
            continue
        }
        switch -regexp -- $active {
            historial|urlHistory {
                 lappend urlHistory $line
            }
            urlIndex|hisIndex {
                set urlIndex $line
            }
            proxy {
                set options(httpProxy) $line
                set options(ftpProxy)  $line
                set options(sameProxy) 1
                set options(proxy) 1
            }
            httpProxy {
                set options(httpProxy) $line
                set options(proxy)     1
            }
            ftpProxy {
                set options(ftpProxy)  $line
                set options(proxy)     1
            }
            sameProxy {
                set options(sameProxy) $line
            }
            ^language$ {
                set options(lang) $line
            }
            scripts {
                set dirGetleft(scripts) $line
            }
            images {
                set dirGetleft(images) $line
            }
            doc {
                set dirGetleft(doc) $line
            }
            languages {
                set dirGetleft(languages) $line
            }
            lastDir {
                set dirGetleft(lastDir) $line
            }
            resumeSites {
                lappend options(resumeSites) $line
            }
        }
    }

    close $handle

    if {[info exists update]} {
        file delete $ficherorc
        catch {file delete [file join $env(HOME) Geterror.log]}
        SaveConfig
    }

    if {![info exists options(lang)]} {
        set options(lang) en
    }
    if {![info exists urlIndex]} {
        set urlIndex 0
    }

    if [catch {return [llength $urlHistory]} ] {
        return 0
    }
}

###############################################################################
# SaveConfig
#   Saves the configuration
###############################################################################
proc SaveConfig {} {
    global urlHistory urlIndex dirGetleft options
    global env tcl_platform
    global labelMessages labelTitles

    if [regexp {Windows} $tcl_platform(os)] {
        set ficherorc [file join "$dirGetleft(conf)" getleft.ini]
    } else {
        set ficherorc [file join "$dirGetleft(conf)" getleft.conf]
    }

    if [catch {open  $ficherorc w+} fichero] {
        tk_messageBox -title $labelTitles(error) -icon Error \
            -message $labelMessages(saveConfig)
        return
    }

    puts $fichero "\[urlHistory\]\n"
    if {![catch {llength $urlHistory} len]} {
        for {set i 0} {$i<$len} {incr i} {
            puts $fichero [lindex $urlHistory $i]
        }
    }

    puts $fichero "\n\[urlIndex\]\n"
    puts $fichero $urlIndex

    if [info exists options(httpProxy)] {
        puts $fichero "\n\[httpProxy\]\n"
        puts $fichero "$options(httpProxy)"
        puts $fichero "\n\[ftpProxy\]\n"
        puts $fichero "$options(ftpProxy)"
        puts $fichero "\n\[sameProxy\]\n"
        puts $fichero "$options(sameProxy)"
    }

    puts $fichero "\n\[language\]\n"
    puts $fichero "$options(lang)"

    puts $fichero "\n\[lastDir\]\n"
    puts $fichero "$dirGetleft(lastDir)"

    if {[llength $options(resumeSites)]} {
        puts $fichero "\n\[resumeSites\]\n"
        foreach site $options(resumeSites) {
            puts $fichero $site
        }
    }

    close $fichero

    return
}

###############################################################################
# PutUrlInHistory
#    Put the given Url in the history of url
#
# Parameter
#    Puts, if need be, the given url in the history of urls.
###############################################################################
proc PutUrlInHistory {url} {
    global urlHistory urlIndex

    set listItems [llength $urlHistory]
    set newListItems [llength $Ccombinado::cbArgs(items)]
    if {$listItems!=$newListItems} {
        set urlHistory $Ccombinado::cbArgs(items)
        set urlIndex   $newListItems
        SaveConfig
    }

    if {![EnHistorial $url]} {
        if {[llength $urlHistory]<10} {
            lappend urlHistory $url
        } else {
            set urlHistory [lreplace $urlHistory $urlIndex $urlIndex $url]
        }
        if {$urlIndex==9} {
            set urlIndex 0
        } else {
            incr urlIndex
        }
        SaveConfig
    }
    return
}

###############################################################################
# MostrarHistorial
#    Shows the url history
###############################################################################
proc MostrarHistorial {} {
    global urlHistory urlIndex

    set len [llength $urlHistory]
    for {set i 0} {$i<$len} {incr i} {
        puts "$i: [lindex $urlHistory $i]"
    }
    return
}

###############################################################################
# EnHistorial
#    Checks whether the given url is already in the url history
#
# Parameter:
#    pagina: given url
#
# Returns:
#    1 if it is.
#    0 if it is not
###############################################################################
proc EnHistorial {pagina} {
    global urlHistory

    set len [llength $urlHistory]
    for {set i 0} {$i<$len} {incr i} {
        if [catch {regsub {(http://)||(ftp://)} \
                [lindex $urlHistory $i] {} una}] {
            return 0
        }
        regsub {(http://)||(ftp://)} $pagina {} dos

        if [string match $una $dos] {
            return 1
        }
    }
    return 0
}

###############################################################################
# ChangeLanguage
#    Changes the default language for the program
#
# Parameters:
#    lang: language to be used
###############################################################################
proc ChangeLanguage {lang} {
    global options

    set options(lang) $lang
    MenusLabels $lang

    SaveConfig

    return
}

###############################################################################
# ReadSupportedLang
#    Reads the 'supported' file to know which languages this version supports
#    and puts then in supportedLang global variable
###############################################################################
proc ReadSupportedLang {} {
    global dirGetleft supportedLang
    global labelTitles

    if [catch {open [file join $dirGetleft(languages) supported] r} handle] {
        tk_messageBox -type ok -icon error -title Error -message $handle
        return
    }

    for {} {![eof $handle]} {} {
        set line [gets $handle]
        if [regexp {^#|^ |^$} $line] {
            continue
        }
        set language [lindex $line 1]
        set key      [lindex $line 0]
        set encoding [lindex $line 2]

        set supportedLang($language) $key
        set supportedLang($key)      $encoding
    }
    close $handle
    return
}

###############################################################################
# ReadLangFile
#    Reads one of the files with translations
#
# Parameter
#    The two letters code that identifies one language, i.e., 'es', 'en'
#
# Returns
#    - '0' all went well
#    - '1' there was an error
###############################################################################
proc ReadLangFile {sufix} {
    global dirGetleft supportedLang
    global labelMenus labelButtons  labelTitles labelMessages
    global labelDialogs labelMonths labelFrames indexMenus

    set languageFile [file join $dirGetleft(languages) menus.$sufix]

    if [catch {open $languageFile r} handle] {
        tk_messageBox -icon error -type ok -title Error \
                -message "File not found"
        return 1
    }
    fconfigure $handle -encoding $supportedLang($sufix)

    for {set active ""} {![eof $handle]} {} {
        set line [gets $handle]
        if [regexp {^#|^ |^$} $line] {
            continue
        }
        if {![string compare $line "End"]} {
            set active ""
            continue
        }
        set key   [lindex $line 0]
        set value [lindex $line 1]

        switch -exact -- $active {
            Menus {
                set indexMenus($key) [string first & $value]
                if {$indexMenus($key)!=-1} {
                    regsub {&} $value {} value
                }
                set labelMenus($key) $value
            }
            Buttons {
                set labelButtons($key) $value
            }
            Titles {
                set labelTitles($key) $value
            }
            Messages {
                set labelMessages($key) $value
            }
            Dialogs {
                set labelDialogs($key) $value
            }
            Months {
                set labelMonths($key) $value
            }
            Frames {
                set labelFrames($key) $value
            }
            "" {
                set active $line
            }
        }
    }
    close $handle
    return 0
}

###############################################################################
# SetEncodings
#    Sets the array that maps the encoding names used in http to the
#    equivalent Tcl encodings.
###############################################################################
proc SetEncodings {} {
    global charSets

    set charSets(ansi_x3.4-1968)        ascii
    set charSets(iso-ir-6)              ascii
    set charSets(iso_646.irv:1991)      ascii
    set charSets(iso646-us)             ascii
    set charSets(us-ascii)              ascii
    set charSets(us)                    ascii
    set charSets(ibm367)                ascii
    set charSets(cp367)                 ascii
    set charSets(ascii)                 ascii

    set charSets(ks_c_5601-1987)        ksc5601
    set charSets(iso-ir-149)            ksc5601
    set charSets(ks_c_5601-1989)        ksc5601
    set charSets(ksc_5601)              ksc5601
    set charSets(korean)                ksc5601

    set charSets(iso-2022-kr)           iso2022-kr

    set charSets(euc-kr)                euc-kr

    set charSets(iso-2022-jp)           iso2022-jp

    set charSets(iso_8859-1:1987)       iso8859-1
    set charSets(iso-ir-100)            iso8859-1
    set charSets(iso_8859-1)            iso8859-1
    set charSets(iso-8859-1)            iso8859-1
    set charSets(latin1)                iso8859-1
    set charSets(l1)                    iso8859-1
    set charSets(ibm819)                iso8859-1
    set charSets(cp819)                 iso8859-1

    set charSets(iso_8859-2:1987)       iso8859-2
    set charSets(iso-ir-101)            iso8859-2
    set charSets(iso_8859-2)            iso8859-2
    set charSets(iso-8859-2)            iso8859-2
    set charSets(latin2)                iso8859-2
    set charSets(l2)                    iso8859-2

    set charSets(iso_8859-3:1988)       iso8859-3
    set charSets(iso-ir-109)            iso8859-3
    set charSets(iso_8859-3)            iso8859-3
    set charSets(iso-8859-3)            iso8859-3
    set charSets(latin3)                iso8859-3

    set charSets(iso_8859-4:1988)       iso8859-4
    set charSets(iso-ir-110)            iso8859-4
    set charSets(iso_8859-4)            iso8859-4
    set charSets(iso-8859-4)            iso8859-4
    set charSets(latin4)                iso8859-4
    set charSets(l4)                    iso8859-4

    set charSets(iso_8859-6:1987)       iso8859-6
    set charSets(iso-ir-127)            iso8859-6
    set charSets(iso_8859-6)            iso8859-6
    set charSets(iso-8859-6)            iso8859-6
    set charSets(ecma-114)              iso8859-6
    set charSets(asmo-708)              iso8859-6
    set charSets(arabic)                iso8859-6

    set charSets(iso_8859-7:1987)       iso8859-7
    set charSets(iso-ir-126)            iso8859-7
    set charSets(iso_8859-7)            iso8859-7
    set charSets(iso-8859-7)            iso8859-7
    set charSets(elot_928)              iso8859-7
    set charSets(ecma-118)              iso8859-7
    set charSets(greek)                 iso8859-7
    set charSets(greek8)                iso8859-7

    set charSets(iso_8859-8:1988)       iso8859-8
    set charSets(iso-ir-138)            iso8859-8
    set charSets(iso_8859-8)            iso8859-8
    set charSets(iso-8859-8)            iso8859-8
    set charSets(hebrew)                iso8859-8

    set charSets(iso_8859-5:1988)       iso8859-5
    set charSets(iso-ir-144)            iso8859-5
    set charSets(iso_8859-5)            iso8859-5
    set charSets(iso-8859-5)            iso8859-5
    set charSets(cyrillic)              iso8859-5

    set charSets(iso_8859-9:1989)       iso8859-9
    set charSets(iso-ir-148)            iso8859-9
    set charSets(iso_8859-9)            iso8859-9
    set charSets(iso-8859-9)            iso8859-9
    set charSets(latin5)                iso8859-9
    set charSets(l5)                    iso8859-9

    set charSets(jis_x0212-1990)        jis0212
    set charSets(x0212)                 jis0212
    set charSets(iso-ir-159)            jis0212

    set charSets(jis_x0201)             jis0201
    set charSets(x0201)                 jis0201

    set charSets(jis_c6226-1983)        jis0208
    set charSets(iso-ir-87)             jis0208
    set charSets(x0208)                 jis0208
    set charSets(jis_x0208-1983)        jis0208

    set charSets(ibm437)                cp437
    set charSets(cp437)                 cp437
    set charSets(437)                   cp437

    set charSets(ibm850)                cp850
    set charSets(cp850)                 cp850
    set charSets(850)                   cp850

    set charSets(ibm852)                cp852
    set charSets(cp852)                 cp852
    set charSets(852)                   cp852

    set charSets(ibm855)                cp855
    set charSets(cp855)                 cp855
    set charSets(855)                   cp855

    set charSets(ibm857)                cp857
    set charSets(cp857)                 cp857
    set charSets(857)                   cp857

    set charSets(ibm860)                cp860
    set charSets(cp860)                 cp860
    set charSets(860)                   cp860

    set charSets(ibm861)                cp861
    set charSets(cp861)                 cp861
    set charSets(861)                   cp861
    set charSets(cp-is)                 cp861

    set charSets(ibm862)                cp862
    set charSets(cp862)                 cp862
    set charSets(862)                   cp862

    set charSets(ibm863)                cp863
    set charSets(cp863)                 cp863
    set charSets(863)                   cp863

    set charSets(ibm864)                cp864
    set charSets(cp864)                 cp864
    set charSets(864)                   cp864

    set charSets(ibm865)                cp865
    set charSets(cp865)                 cp865
    set charSets(865)                   cp865

    set charSets(ibm869)                cp869
    set charSets(cp869)                 cp869
    set charSets(869)                   cp869
    set charSets(cp-gr)                 cp869

    set charSets(koi8-r)                koi8-r

    return
}

###############################################################################
# SetDefaultOptions
#    Sets the default options.
###############################################################################
proc SetDefaultOptions {} {
    global options

    set options(dir)             1
    set options(cgi)             0
    set options(resume)          1
    set options(update)          0
    set options(external)        0
    set options(levels)         -1
    set options(exLevels)        0
    set options(images)          0
    set options(filter)         ""
    set options(map)             0

    return
}

###############################################################################
# SetResumeSitesMenu
#    Puts the entries in the "Resume Sites" menu.
###############################################################################
proc SetResumeSitesMenu {} {
    global options

    .menus.fichero.resume delete 0 end

    foreach site [lsort $options(resumeSites)] {
        .menus.fichero.resume add command -label $site \
                -command "AutomaticDownload $site"
    }
    return
}

###############################################################################
# MenusLabels
#    Puts the labels into the menus.
#
# Parameters:
#    labFile: file with the labels in the, hopefully, desired language
###############################################################################
proc MenusLabels {{labFile ""}} {
    global menus dirGetleft
    global labelButtons labelTitles labelMessages labelDialogs labelMonths
    global labelMenus indexMenus

    if {$labFile==""} {
        set labFile en
    }

    ReadLangFile $labFile

    .menus      entryconfigure 1 -label $labelMenus(file)           \
            -underline $indexMenus(file)
    $menus(1)   entryconfigure 0 -label $labelMenus(enterUrl)       \
            -underline $indexMenus(enterUrl)
    $menus(1)   entryconfigure 1 -label $labelMenus(siteMap)        \
            -underline $indexMenus(siteMap)
    $menus(1)   entryconfigure 2 -label $labelMenus(resumeSite)     \
            -underline $indexMenus(resumeSite)
    $menus(1)   entryconfigure 4 -label $labelMenus(stop)           \
            -underline $indexMenus(stop)
    $menus(1,4) entryconfigure 0 -label $labelMenus(stopPage)       \
            -underline $indexMenus(stopPage)
    $menus(1,4) entryconfigure 1 -label $labelMenus(stopFile)       \
            -underline $indexMenus(stopFile)
    $menus(1,4) entryconfigure 2 -label $labelMenus(stopNow)        \
            -underline $indexMenus(stopNow)
    $menus(1)   entryconfigure 5 -label $labelMenus(pause)          \
            -underline $indexMenus(pause)
    $menus(1,5) entryconfigure 0 -label $labelMenus(stopPage)       \
            -underline $indexMenus(stopPage)
    $menus(1,5) entryconfigure 1 -label $labelMenus(stopFile)       \
            -underline $indexMenus(stopFile)
    $menus(1)   entryconfigure 7 -label $labelMenus(exit)           \
            -underline $indexMenus(exit)

    .menus      entryconfigure 2 -label $labelMenus(options)        \
            -underline $indexMenus(options)
    $menus(2)   entryconfigure 0 -label $labelMenus(all)            \
            -underline $indexMenus(all)
    $menus(2)   entryconfigure 1 -label $labelMenus(uplinks)        \
            -underline $indexMenus(uplinks)
    $menus(2,0) entryconfigure 0 -label $labelMenus(linksFollow)    \
            -underline $indexMenus(linksFollow)
    $menus(2,0) entryconfigure 1 -label $labelMenus(linksIgnore)    \
            -underline $indexMenus(linksIgnore)
    $menus(2)   entryconfigure 2 -label $labelMenus(levels)         \
            -underline $indexMenus(levels)
    $menus(2,0) entryconfigure 2 -label $labelMenus(linksIgnore)    \
            -underline $indexMenus(linksIgnore)
    $menus(2,1) entryconfigure 0 -label $labelMenus(noLimit)        \
            -underline $indexMenus(noLimit)
    $menus(2)   entryconfigure 3 -label $labelMenus(external)       \
            -underline $indexMenus(external)
    $menus(2,2) entryconfigure 0 -label $labelMenus(linksIgnore)    \
            -underline $indexMenus(linksIgnore)
    $menus(2)   entryconfigure 4 -label $labelMenus(fileFilter)     \
            -underline $indexMenus(fileFilter)
    $menus(2,3) entryconfigure 0 -label $labelMenus(onlyHtml)       \
            -underline $indexMenus(onlyHtml)
    $menus(2,3) entryconfigure 1 -label $labelMenus(images)         \
            -underline $indexMenus(images)
    $menus(2,3,1) entryconfigure 0 -label $labelMenus(allImages)    \
            -underline $indexMenus(allImages)
    $menus(2,3,1) entryconfigure 1 -label $labelMenus(onlyThumb)    \
            -underline $indexMenus(onlyThumb)
    $menus(2,3,1) entryconfigure 2 -label $labelMenus(noThumb)      \
            -underline $indexMenus(noThumb)
    $menus(2,3) entryconfigure 2 -label $labelMenus(chooseFilter)   \
            -underline $indexMenus(chooseFilter)
    $menus(2)   entryconfigure 5 -label $labelMenus(resume)         \
            -underline $indexMenus(resume)
    $menus(2)   entryconfigure 6 -label $labelMenus(update)         \
            -underline $indexMenus(update)
    $menus(2)   entryconfigure 7 -label $labelMenus(cgi)            \
            -underline $indexMenus(cgi)
    $menus(2)   entryconfigure 8 -label $labelMenus(useProxy)       \
            -underline $indexMenus(useProxy)

    .menus      entryconfigure 3 -label $labelMenus(tools)          \
            -underline $indexMenus(tools)
    $menus(3)   entryconfigure 0 -label $labelMenus(purgeFiles)     \
            -underline $indexMenus(purgeFiles)
    $menus(3)   entryconfigure 1 -label $labelMenus(restoreOrig)    \
            -underline $indexMenus(restoreOrig)
    $menus(3)   entryconfigure 2 -label $labelMenus(updateFiles)    \
            -underline $indexMenus(updateFiles)
    $menus(3)   entryconfigure 3 -label $labelMenus(configureProxy) \
            -underline $indexMenus(configureProxy)
    $menus(3)   entryconfigure 4 -label $labelMenus(languages)      \
            -underline $indexMenus(languages)
    if {$::DEBUG==1} {
        $menus(3)   entryconfigure 5 -label $labelMenus(resource)   \
                -underline $indexMenus(resource)
    }

    .menus      entryconfigure 4 -label $labelMenus(help)           \
            -underline $indexMenus(help)
    $menus(4)   entryconfigure 0 -label $labelMenus(manual)         \
            -underline $indexMenus(manual)
    $menus(4)   entryconfigure 2 -label $labelMenus(changes)        \
            -underline $indexMenus(changes)
    $menus(4)   entryconfigure 3 -label $labelMenus(license)        \
            -underline $indexMenus(license)
    $menus(4)   entryconfigure 5 -label $labelMenus(about)          \
            -underline $indexMenus(about)

    return
}

###############################################################################
# ShowMenu1
#    Before posting the file menu, this procedure decides which entries
#    should be active, and which will have to wait
###############################################################################
proc ShowMenu1 {} {
    global getleftState options

    if {$getleftState(downloading)==0} {
        .menus.fichero entryconfigure 0 -state normal
        .menus.fichero entryconfigure 1 -state normal
        .menus.fichero entryconfigure 2 -state normal
        .menus.fichero entryconfigure 4 -state disabled
        .menus.fichero entryconfigure 5 -state disabled
        if {[llength $options(resumeSites)]==0} {
            .menus.fichero entryconfigure 2 -state disabled
        }
    } else {
        .menus.fichero entryconfigure 0 -state disabled
        .menus.fichero entryconfigure 1 -state disabled
        .menus.fichero entryconfigure 2 -state disabled
        .menus.fichero entryconfigure 4 -state normal
        .menus.fichero entryconfigure 5 -state normal
    }

    return
}

###############################################################################
# CreateMenus
#   Creates the menus, without putting the labels
###############################################################################
proc CreateMenus {} {
    global options dirGetleft menus supportedLang

    menu .menus -relief flat

    set menus(1) [menu .menus.fichero      -tearoff 0 -postcommand ShowMenu1]
    .menus add cascade -menu $menus(1)
    set menus(2) [menu .menus.editar       -tearoff 0]
    .menus add cascade -menu $menus(2)
    set menus(3) [menu .menus.herramientas -tearoff 0]
    .menus add cascade -menu $menus(3)
    set menus(4) [menu .menus.help         -tearoff 0]
    .menus add cascade -menu $menus(4)

    $menus(1) add command -command EnterUrl
    $menus(1) add command -command GetSiteMap
    $menus(1) add cascade -menu $menus(1).resume
    $menus(1) add separator
    $menus(1) add cascade -menu $menus(1).stop
    $menus(1) add cascade -menu $menus(1).pause
    $menus(1) add separator
    $menus(1) add command -command exit

    set menus(1,2) [menu $menus(1).resume -tearoff 0]
    SetResumeSitesMenu

    set menus(1,4) [menu $menus(1).stop -tearoff 0]
    $menus(1,4) add check -variable options(stopPage)  -onvalue 1 -offvalue 0
    $menus(1,4) add check -variable options(stopFile)  -onvalue 1 -offvalue 0
    $menus(1,4) add command -command ::Ventana::CancelDownloading

    set menus(1,5) [menu $menus(1).pause -tearoff 0]
    $menus(1,5) add check -variable options(pausePage) -onvalue 1 -offvalue 0
    $menus(1,5) add check -variable options(pauseFile) -onvalue 1 -offvalue 0

    $menus(2) add command -command ::Opciones::ChooseOptions
    $menus(2) add cascade -menu $menus(2).directorios
    $menus(2) add cascade -menu $menus(2).levels
    $menus(2) add cascade -menu $menus(2).exLevels
    $menus(2) add cascade -menu $menus(2).filters
    $menus(2) add check   -variable options(resume)    -onvalue 1 -offvalue 0
    $menus(2) add check   -variable options(update)    -onvalue 1 -offvalue 0
    $menus(2) add check   -variable options(cgi)       -onvalue 1 -offvalue 0 \
        -command {catch {unset urlsDownloaded}}
    $menus(2) add check   -variable options(proxy)     -onvalue 1 -offvalue 0

    set menus(2,0) [menu $menus(2).directorios -tearoff 0]
    $menus(2,0) add radio -variable options(dir) -value 1 \
            -command {catch {unset urlsDownloaded}}
    $menus(2,0) add radio -variable options(dir) -value 0 \
            -command {catch {unset urlsDownloaded}}

    set menus(2,1) [menu $menus(2).levels -tearoff 0]
    $menus(2,1) add radio -variable options(levels) -value -1
    for {set i 0} {$i<6} {incr i} {
        $menus(2,1) add radio -label $i -variable options(levels) -value $i \
                -command {catch {unset urlsDownloaded}}
    }

    set menus(2,2) [menu $menus(2).exLevels -tearoff 0]
    $menus(2,2) add radio -variable options(exLevels) -value 0      \
            -command {
                set options(external) 0
                catch {unset urlsDownloaded}
            }
    for {set i 1} {$i<4} {incr i} {
        $menus(2,2) add radio -label $i -variable options(exLevels) \
            -value $i -command {
                set options(external) 1
                catch {unset urlsDownloaded}
            }
    }

    set menus(2,3) [menu $menus(2).filters -tearoff 0]
    $menus(2,3) add check -variable options(onlyHtml) -onvalue 1 -offvalue 0  \
                -command {catch {unset urlsDownloaded}}
    $menus(2,3) add cascade -menu $menus(2,3).images
    $menus(2,3) add command -command {
                Herramientas::FilterFiles
            }
    set menus(2,3,1) [menu $menus(2,3).images -tearoff 0]
    $menus(2,3,1) add radio -variable options(images) -value 0 \
            -command {catch {unset urlsDownloaded}}
    $menus(2,3,1) add radio -variable options(images) -value 1 \
            -command {catch {unset urlsDownloaded}}
    $menus(2,3,1) add radio -variable options(images) -value 2 \
            -command {catch {unset urlsDownloaded}}

    $menus(3) add command -command Herramientas::PurgeFiles
    $menus(3) add command -command Herramientas::RestoreOriginals
    $menus(3) add command -command Herramientas::UpdateFiles
    $menus(3) add command -command Herramientas::ConfProxy
    $menus(3) add cascade -menu $menus(3).idiomas

    if {$::DEBUG==1} {
        $menus(3) add command -command \
                "source [file join "$dirGetleft(main)" Getleft.tcl]"
    }

    set menus(3,0) [menu $menus(3).idiomas -tearoff 0]
    set i 0
    foreach lang [lsort [array names supportedLang]] {
        if [file exists [file join $dirGetleft(languages) \
                menus.$supportedLang($lang)]] {
            $menus(3,0) add command \
                    -command "ChangeLanguage $supportedLang($lang)"
            $menus(3,0) entryconfigure $i -label $lang
            incr i
        }
    }

    $menus(4) add command -command Ayuda::Manual
    $menus(4) add separator
    $menus(4) add command -command Ayuda::Changes
    $menus(4) add command -command Ayuda::Licence
    $menus(4) add separator
    $menus(4) add command -command Ayuda::About

    MenusLabels $options(lang)

    return
}

###############################################################################
#				PROGRAMA PRINCIPAL
###############################################################################

set DEBUG 0

wm withdraw .

CheckPrograms

set dirGetleft(main) [file dirname [info script]]
if [file exists /etc/getlet.conf] {
    ReadConfig /etc/getleft.conf
} elseif [file exists /usr/local/etc/getleft.conf] {
    ReadConfig /usr/local/etc/getleft.conf
} else {
    regexp {(?:.:)(.*)} "$dirGetleft(main)" nada dirGetleft(main)
    set dirGetleft(scripts)   [file join "$dirGetleft(main)" scripts]
    set dirGetleft(images)    [file join "$dirGetleft(main)" images]
    set dirGetleft(doc)       [file join "$dirGetleft(main)" doc]
    set dirGetleft(languages) [file join "$dirGetleft(main)" Languages]
}

if [regexp {Windows} $tcl_platform(os)] {
    set dirGetleft(conf) $dirGetleft(main)
} else {
    set dirGetleft(conf) [file join $env(HOME) .getleft]
}

if {![file exists $dirGetleft(conf)]} {
    if [catch {file mkdir $dirGetleft(conf)} error] {
        tk_messageBox -title Error -icon error -type ok \
                -message "Couldn't create dir:\n$error"
        exit
    }
}

# First I named these files with the 'opt' extension only to
# change my mind and use 'gdt', unfortunately I get to
# maintain the first extension for life

set fileList [glob -nocomplain [file join $dirGetleft(conf) *.opt]]
foreach file $fileList {
    set root [file rootname $file]
    file rename -force -- $file $root.gdt
}

source [file join "$dirGetleft(scripts)" Ccombinado.tcl]
source [file join "$dirGetleft(scripts)" Ventana.tcl]
source [file join "$dirGetleft(scripts)" Dialogos.tcl]
source [file join "$dirGetleft(scripts)" Herramientas.tcl]
source [file join "$dirGetleft(scripts)" Ayuda.tcl]
source [file join "$dirGetleft(scripts)" HtmlParser.tcl]
source [file join "$dirGetleft(scripts)" balloon.tcl]
source [file join "$dirGetleft(scripts)" Commands.tcl]
source [file join "$dirGetleft(scripts)" FrameLabel.tcl]
source [file join "$dirGetleft(scripts)" Opciones.tcl]
source [file join "$dirGetleft(scripts)" CuadroSpin.tcl]
source [file join "$dirGetleft(scripts)" Descargar.tcl]
source [file join "$dirGetleft(scripts)" Progreso.tcl]
source [file join "$dirGetleft(scripts)" Cookies.tcl]

if {![winfo exists .menus]} {
    ReadSupportedLang
    ReadLangFile en
    SetEncodings
    SetDefaultOptions
    Ventana::Rizo::SetCurlVersion

    wm geometry . 300x1
    wm title . "Getleft v0.10.1"

    . config -menu .menus

    wm protocol . WM_DELETE_WINDOW {
        SaveConfig
        exit
    }

    ReadConfig

    bind Button <Return>   {%W invoke}
    bind Button <KP_Enter> {%W invoke}

    CreateMenus
    set getleftState(downloading)      0
    set getleftState(mapDone)          0
}
wm deiconify .

