###############################################################################
###############################################################################
##                               Descargar.tcl
###############################################################################
###############################################################################
## Includes the procedures needed for downloading a link
###############################################################################
###############################################################################
## (c) 2001 Andrs Garca Garca. fandom@retemail.es
## You may distribute the contents of this file under the terms of the GPL v2
###############################################################################
###############################################################################

namespace eval Descargar {

###############################################################################
# FileExists
#    Checks whether the file we are going to download has already been 
#    downloaded.
#
# Parameter:
#    link: The url about to be downloaded
#    fileName: Name of the file to check
#    sufix: '1' if the fileName ends in 'html?' '0' otherwise.
#
# Returns:
#    - '-1' the file does not exist.
#    - '0' file exists and is not a Html page.
#    - '1' file exists and is a Html page
#    - '4' if it exists and the referrer page will have to change for it.
###############################################################################
proc FileExists {link fileName sufix} {
    global options
    variable updated
    variable relocated

    if {[file exists $fileName]} {
        if {($options(update)==0)||([info exists "updated($fileName)"])} {
            if {$sufix==0} {
                if {([file isdirectory $fileName])&&(![regexp {/$} $fileName])} {
        	        if [file exists $fileName/index.html] {
            	        set relocated(url) $link/index.html
                	   return 4
                    }
                } elseif [file exists $fileName.html] {
                    return 1
                } else {
                    return 0
                }
            } else {
                return 1
            }
        }
    }
    return -1
}

###############################################################################
# UpdateFile
#    Checks whether the file in the server is newer than the one we have.
#
# Parameter:
#    fileName: Name of the file to check
#
# Returns:
#    - '1' if we have to update
#    - '0' if not
###############################################################################
proc UpdateFile {fileName} {
    global ::Ventana::Rizo::meta
    global getleftState options
    global labelTitles labelMessages

    if [file exists $fileName.orig] {
        set oldFile [file mtime $fileName.orig]
    } else {
        set oldFile [file mtime $fileName]
    }
    if {[catch "set meta(lastModified)"]} {
        if {$getleftState(filesChosen)==0} {
            set which [tk_messageBox -icon info -type yesno \
                    -title $labelTitles(error) \
                    -message $labelMessages(noDate)]
            if {$which=="no"} {
                set options(stopFile) 1
                return 0
            }
        }
        set lastChange 0
    } else {
        set lastChange [clock scan $meta(lastModified)]
    }
    if {$lastChange<$oldFile} {
        return 0
    }
    return 1
}

###############################################################################
# DownloadHead
#    Downloads the headers of the link
#
# Parameter:
#    link:    Link to download
#    referer: Referrer page
#    sufix: '1' if the link has a 'html' extension.
#
# Returns:
#    - '4' in case of a relocation and a HTML page
#    - '3' in case of a relocation and not a HTML page
#    - '2' in case of error
#    - '1' if it is a HTML page
#    - '0' if it is not
###############################################################################
proc DownloadHead {link referer sufix} {
    global Ventana::Rizo::meta
    global siteUrl
    global errorCode
    variable filesNotFound
    variable relocated

    for {set relocated(ok) 0} {1==1} {} {
        if {[regexp {^ftp://} $link]} {
            set meta(content) "nada"
            set meta(totalBytes) 0
            break
        }
        Ventana::HeadDownloading $link $referer
        if {($::Ventana::Rizo::curlError!=0)||($::Ventana::Rizo::errorMessage!="")} {
            ::Herramientas::ErrorLog $link $referer
            if {[lindex $errorCode 2]>=400} {
                set filesNotFound($link) 1
            }
            if {$::DEBUG==1} {
                puts "curlError: $::Ventana::Rizo::curlError: $link"
            }
            if {[lindex $errorCode 2]!=550} {	    
                return 2
            }	    
        }
        if {$meta(relocate)!=""} {
            if {$::DEBUG==1} {
                puts "La relocalizacin: $meta(relocate)"
            }
            set relocated(url) $meta(relocate)
            if {([regexp {\?} $meta(relocate)])} {
                return 2
            }
            set parsedUrl [HtmlParser::ParseUrl $meta(relocate)]
            if {$parsedUrl==1} {
                return 2
            }
            set newWWW      [lindex $parsedUrl 1]
            set newDir      [lindex $parsedUrl 2]
            set newFileName [lindex $parsedUrl 3]
            if {[string compare [string tolower $newWWW] \
                                [string tolower $siteUrl(www)]]} {
                set Ventana::Rizo::errorMessage "Relocated link\n$meta(relocate)"
                return 2
            }
            if {$newFileName==""} {
                set newFileName index.html
            }
            set newFileName [TidyNames [file join $newDir $newFileName]]
            set fileName    [UrlToFile $meta(relocate)]
            set link        $meta(relocate)
            set relocated(ok) 1
            if {[file exists $fileName]} {
                if {([regexp -nocase {\.html?$} $fileName])||\
                    ([file exists $fileName.html])} {
                    set isHtml 1
                } else {
                    set isHtml 0
                }
                return [expr 3 + $isHtml]
            }
        } else {
            break
        }
    }
# Some servers don't return a mime time, that's why we also check the sufix
    if {($meta(content)!="text/html")&&($sufix!=1)} { 
        return 0
    }
    return 1
}


###############################################################################
# DownloadFile
#    Downloads the link
#
# Parameter:
#    link:     Link to download
#    referer:  referrer page
#    fileName: file is which the link will be saved.
#
# Returns:
#    - '1' in case of error
#    - '0' all is well
###############################################################################
proc DownloadFile {link referer fileName} {
    global errorCode labelMessages

    set Ventana::Rizo::curlReport(parar)    0
    set Ventana::Rizo::curlReport(nextFile) 0
    Ventana::FileDownloading $fileName.$labelMessages(downloading) \
            $link $referer
    tkwait variable Ventana::Rizo::curlReport(nextFile)

    if {$errorCode!=""} {
        if {$::DEBUG==1} {
            puts "Hubo un error: $errorCode: $::errorInfo"
        }
        ::Herramientas::ErrorLog $link $referer
        return 1
    }
    return 0
}

###############################################################################
# Download
#    Downloads a link
#
# Parameter:
#    link:     Link to download
#    referer:  referrer page
#    sufix:    Whether the file ends in 'htm?' (1) or not (0)
#
# Returns:
#    - '4' in case of a relocation and a HTML page
#    - '3' in case of a relocation and not a HTML page
#    - '2' in case of error
#    - '1' if it is a HTML page
#    - '0' if it is not
###############################################################################
proc Download {link referer sufix} {
    global siteUrl options errorCode getleftState
    global Ventana::Rizo::meta
    global directories
    global labelDialogs labelMessages
    variable updated
    variable filesNotFound
    variable relocated

    if {[info exists filesNotFound($link)]} {
        return 2
    }

    set fileName [UrlToFile $link]

    set returnCode [FileExists $link $fileName $sufix]
    if {$returnCode!=-1} {
        return $returnCode
    } else {
        if {$options(update)==1} {
            set updated($fileName) 1
        }
    }

    set returnCode [DownloadHead $link $referer $sufix]
    switch -regexp -- $returnCode {
        {0|1}   "set isHtml $returnCode"
        {2|3|4} "return $returnCode"
    }        

    if {$options(cancelDown)} {
        return 2
    }

    if {$relocated(ok)==1} {
        set link $relocated(url)
        set fileName [UrlToFile $link]
    }    

    if {($isHtml==0)&&($options(onlyHtml)==1)} {
        return 0
    }

    if {($getleftState(filesChosen)==0)&&($isHtml==0)} {
        set fileName $directories(base)/[file tail $fileName]
    }

    set returnCode [CheckDir $fileName]
    if {$returnCode==1} {
        return 2
    }

    regexp {(.*)(://)} $link nada prot
    if {($prot!="ftp")&&($meta(totalBytes)==0)} {
        catch {error "No content" GETLEFT \
                "Getleft \"$labelDialogs(noContent)\" 002"}
        Herramientas::ErrorLog $link $referer
        return 2
    }

    if {($options(update)==1)&&([file exists $fileName])} {
        if {![UpdateFile $fileName]} {
            return $isHtml
        }
    }

    if [DownloadFile $link $referer $fileName] {
        return 2
    }            

    catch {file rename -force $fileName.$labelMessages(downloading) $fileName}
    if {$isHtml==1} {
        file delete $fileName.orig
    }

    if {$relocated(ok)==1} {
        return [expr 3 + $isHtml]
    }
    return $isHtml
}

###############################################################################
# DownloadLinks
#    Download all the links taken from a referrer page.
#
# Parameters:
#    urlList: list with the urls to be downloaded.
#    referer: html page from which the links come from.
#
# Returns
#    The link list with only the Html links.
###############################################################################
proc DownloadLinks {urlList referer} {
    global options
    variable relocated

    set invokeSed 0

    set urlListTemp ""
    foreach urlToDownload $urlList {
        update
        if {$options(onlyHtml)==1} {
            if [regexp -nocase -expanded {(jpg)|(jpeg)|(gif)|(gz)|(tar)|(zip)|\
                                          (exe)|(ps)|(pdf)|(txt)|(text)|(avi)|\
                                          (ram)|(wav)|(png)|(tif)|(mov)|(qt)} \
                [file extension $urlToDownload]] {
                continue
            }
        }
        set fileName [UrlToFile $urlToDownload]

        set changedPageLink [RelativePath $referer $urlToDownload]
        regsub -all {&} $changedPageLink {\\\&} changedAmper

        if [regexp {/$} $urlToDownload] {
            set urlListTemp [concat $urlListTemp ${urlToDownload}index.html]
        }

        if [regexp -nocase {\.html?} [file extension $fileName]] {
            set sufix 1
        } else {
            set sufix 0
        }

        set isHtml [Download $urlToDownload $referer $sufix]
        switch -regexp -- $isHtml {
            1 {
                if ($sufix==0) {
                    catch {file rename $fileName $fileName.html}
                    Commands::Touch $fileName
                    set urlListTemp [concat $urlListTemp $urlToDownload.html]
                    Commands::SedChangeEnter "(\")($changedPageLink)(\")" \
                                             "\"${changedAmper}.html\""
                    set invokeSed 1
                } else {
                    set urlListTemp [concat $urlListTemp $urlToDownload]
                }
            }
            {3|4} {
                set newLink [RelativePath $referer $relocated(url)]
                set sufix [regexp -nocase {\.html?} [file extension $newLink]]
                if {$isHtml==4} {
                    if {$sufix==0} {
                        set urlListTemp [concat $urlListTemp $relocated(url).html]
                        set newLink $newLink.html
                    } else {
                        set urlListTemp [concat $urlListTemp $relocated(url)]
                    }
                }
                Commands::SedChangeEnter "(\")($changedPageLink)(\")" "\"$newLink\""
                set invokeSed 1
            }
        }
        if {$options(stopFile)==1} {
            break
        }
        if {$options(pauseFile)==1} {
            Ventana::Pause file
        }
    }
    if {$invokeSed==1} {
        Commands::Sed [UrlToFile $referer]
    }
    return $urlListTemp
}

###############################################################################
# ProcessUrlList
#    After all links from a referer page have been downloaded,this checks 
#    for the Html pages between them to continue the download of the site.
#
# Parameters:
#    urlList: list with the urls to be processed fror links
#    level: current level of recursion for internal links.
#    externalLevel: current level of recursion for external link, defaults to 0
###############################################################################
proc ProcessUrlList {urlList level {externalLevel 0}} {
    global siteUrl directories urlsDownloaded options filesDone

    foreach url $urlList {
        set fileToProcess   [UrlToFile $url]
        set parsedUrl       [HtmlParser::ParseUrl $url]
        if {[info exists filesDone($fileToProcess)]} {
            if {($level>=$filesDone($fileToProcess))||($options(levels)==-1)} {
                continue
            }
        }
        if {[info exists urlsDownloaded($url)]} {
            if {($level>=$urlsDownloaded($url))||($options(levels)==-1)} {
                continue
            }
        }
        set tmpUrl        $siteUrl(dir)
        set tmpLocal      $directories(local)
        set tmpProt       $siteUrl(prot)
        set domain        [lindex $parsedUrl 1]
        set siteUrl(dir)  [lindex $parsedUrl 2]
        set siteUrl(prot) [lindex $parsedUrl 0]
        if {[string compare [string tolower $domain] \
                [string tolower $siteUrl(www)]]} {
            set nextExternalLevel [expr $externalLevel + 1]
        } else {
            set nextExternalLevel 0
        }
        if {($level<$options(levels))||($options(levels)==-1)} {
            if {$nextExternalLevel<=$options(exLevels)} {
                Preprocessing $url $level
                if {$HtmlParser::nLinks!=1} {
                    set nextLevel [expr $level +1]
                    ControlDownloading $url $nextLevel $nextExternalLevel
                    if {($options(stopFile)==1)||($options(stopPage)==1)} {
                        return
                    }
                    set urlsDownloaded($url) $level
                }
            }
        }
        set siteUrl(prot)      $tmpProt
        set siteUrl(dir)       $tmpUrl
        set directories(local) $tmpLocal
    }
    return
}

###############################################################################
# PrepareDownloading
#    Prepares the list of files to download
#
# Parameter
#    currentLevel: the current level of recursion
#
# Returns
#    The list with the urls to be downloaded.
###############################################################################
proc PrepareDownloading {currentLevel} {
    global siteIndex siteMap options getleftState

    for {set i 1 ; set urlList ""} {$i<$HtmlParser::nLinks} {incr i} {
        if {$HtmlParser::links($i,ok)==1} {
            lappend urlList $HtmlParser::links($i,url)
            if {$options(map)==1} {
                if {![info exists siteMap($HtmlParser::links($i,url))]} {
                    set siteMap($siteIndex,0) $currentLevel
                    set siteMap($siteIndex,1) $HtmlParser::links($i,url)
                    set siteMap($siteIndex,2) $HtmlParser::links($i,descrip)
                    set siteMap($HtmlParser::links($i,url)) 1
                    incr siteIndex
                }
            }
        }
    }
    if {$getleftState(filesChosen)==0} {
        set getleftState(filesChosen) 1
    }
    return $urlList
}

###############################################################################
# ControlDownloading
#    Send 'Downloading' all the links one by one.
#
# Parameters:
#    referer: html page from which the links come from.
#    level: current level of recursion for internal links.
#    externalLevel: current level of recursion for external link, defaults to 0
###############################################################################
proc ControlDownloading {referer level {externalLevel 0}} {
    global options
    variable relocated 

#    if {$::DEBUG==1} {
#        puts "En control"
#        puts "ListaEnlaces: $urlList"
#    }

    set urlList [PrepareDownloading $level]
    if {[llength $urlList]==0} return

    set urlList [DownloadLinks $urlList $referer]

    if {$options(pausePage)==1} {
        Ventana::Pause page
    }

    if {($options(stopFile)==1)||($options(stopPage)==1)} {
        return
    }

    ProcessUrlList $urlList $level $externalLevel

    return
}

}
