###############################################################################
###############################################################################
#####                            Ventana.tcl
###############################################################################
##### Creates the window which will provide some information to the user 
##### while a file is being downloaded.
###############################################################################
##### Copyright 1999-2000 Andrs Garca Garca - fandom@retemail.es
##### Distributed under the terms of the GPL v2
###############################################################################

namespace eval Ventana {

source [file join "$dirGetleft(scripts)" Rizo.tcl]

###############################################################################
# PauseDownloading
#    Pauses the downloading according to the menu entries.
#
# Parameter
#    which: either 'file' or 'page'
###############################################################################
proc Pause {which} {
    global options labelTitles labelMessages

    set windowState [wm state .bajando]
    if {$windowState=="iconic"} {
        wm deiconify .bajando
    }
    wm withdraw .bajando
    tk_messageBox -title $labelTitles(paused) -icon info -type ok \
         -message $labelMessages(paused)
    if {$which=="file"} {
        tkwait variable options(pauseFile)
    } else {
        tkwait variable options(pausePage)
    }
    wm deiconify .bajando
    if {$windowState=="iconic"} {
        catch {destroy .balloon}
        wm iconify .bajando
    }

    return
}

###############################################################################
# CancelDownloading
#    This procedure is invoked from the downloading window when you click on
#    the 'cancel' button or you want to close the window
#
# Parameter:
#    skip: if skip is 'skip' the current file will be skipped and the next one
#    will begin to be downloaded. If there is no parameter, the downloading
#    will be stopped
###############################################################################
proc CancelDownloading {{skip {} }} {
    global labelTitles labelMessages
    global getleftState
    global ::Ventana::Rizo::curlReport
    variable afterId
    variable window

    set what [tk_messageBox -type yesno -icon question \
            -title $labelTitles(confirm) -message $labelMessages(stopDownload)]

    if {$what=="no"} {
        return
    }
    catch {after cancel $afterId}
    catch {Rizo::Lector $::Ventana::Rizo::pipe stopNow}

    if {($skip=="")||($getleftState(filesChosen)==0)} {
        set ::options(stopFile) 1
    }
    set ::options(cancelDown)   1
    set Ventana::Rizo::curlReport(end) 1

    catch {incr curlReport(nextFile)}

    return
}

###############################################################################
# TrimString
#    Trims the given string so that if it is longer than the available space
#    the middle of it get substituted with '...', and puts the resulting
#    string in the given label.
#
# Parmetros
#    cadena: string to trim.
#    labelPath: the label in which the string is going to be shown
###############################################################################
proc TrimString {cadena labelPath} {

    set font     [$labelPath cget -font]
    set color    [$labelPath cget -bg]
    set tmpColor [$labelPath cget -fg]
    set varName  [$labelPath cget -textvariable]

    upvar 0 $varName textVariable
    $labelPath configure -fg $color

    set textVariable $cadena
    set labelLength  [expr [winfo width $labelPath]-5]
    set strLength    [font measure $font -displayof $labelPath $cadena]

    if {$labelLength<=$strLength} {
        set strChars [string length $cadena]
        set middle   [expr round($strChars/2)-2]
        set primer   [string range $cadena 0 $middle]
        set segun    [string range $cadena [expr $middle+1] end]
        set textVariable ""
        append textVariable $primer "..." $segun
        while {[font measure $font -displayof $labelPath $textVariable]>$labelLength} {
            regsub {(.)(\.\.\.)(.)} $textVariable {...} textVariable
        }
        BalloonHelp::set_balloon $labelPath $cadena
    } else {
        BalloonHelp::delete_balloon $labelPath
    }

    $labelPath configure -fg $tmpColor

    return
}

###############################################################################
# FileStrings
#    Controls the strings that will be put in the downloading window, to
#    show the name and directory of the file being processed or downloaded
#
# Parameter:
#    url to be downloaded or preprocessed
###############################################################################
proc FileStrings {url} {
    global directories labelMessages
    variable window

    set fileName  [UrlToFile $url]
    set parsedUrl [HtmlParser::ParseUrl $url]
    set domain    [lindex $parsedUrl 1]

    regexp "(.*)(\.$labelMessages(downloading))" $fileName nada fileName
    regexp "(?:$directories(base))(.*)" $fileName nada tmp
    regexp {(.*)(?:/)(.*)} $tmp nada tmpDir tmpFile

    TrimString $domain  $window(url)
    TrimString $tmpDir  $window(dir)
    TrimString $tmpFile $window(file)

    update

    return
}

###############################################################################
# DownloadWindow
#    Creates the window, in which some info about the downloading will be
#    displayed.
###############################################################################
proc DownloadWindow {} {
    global labelButtons labelDialogs
    global tcl_platform dirGetleft
    variable window

    set coord(x) [winfo rootx .]
    set coord(y) [winfo rooty .]

    set win [toplevel .bajando]
    wm geometry $win 290x160+[expr $coord(x)+200]+[expr $coord(y)+15]
    wm resizable $win 0 0

    set marco     [frame $win.marco]
    set marco_in  [frame $marco.marco_interno]
    set marco_url [frame $marco_in.url]
    set marco_dir [frame $marco_in.dir]
    set marco_fil [frame $marco_in.fil]
    set marco_que [frame $marco_in.que]
    set marco_bar [frame $marco_in.bar]
    set botones   [frame $win.botones]

    set imaUrl    [image create photo \
            -file [file join "$dirGetleft(images)" html.gif]]
    set imaFolder [image create photo \
            -file [file join "$dirGetleft(images)" folder.gif]]
    set imaFiles  [image create photo \
            -file [file join "$dirGetleft(images)" files.gif]]
    set imaClock  [image create photo \
            -file [file join "$dirGetleft(images)" clock.gif]]

    set etq_url   [label $marco_url.etq_url  -image $imaUrl]
    set etq_dir   [label $marco_dir.etq_dir  -image $imaFolder]
    set etq_file  [label $marco_fil.etq_file -image $imaFiles]
    set etq_que   [label $marco_que.etq_que  -image $imaClock]

    BalloonHelp::set_balloon $etq_url  $labelDialogs(url)
    BalloonHelp::set_balloon $etq_dir  $labelDialogs(dir)
    BalloonHelp::set_balloon $etq_file $labelDialogs(file)
    BalloonHelp::set_balloon $etq_que  $labelDialogs(left)

    set etq_url2  [label $marco_url.etq_url2  -width 45 -anchor w \
            -bd 2 -relief sunken -textvariable Ventana::window(labelUrl)]
    set etq_dir2  [label $marco_dir.etq_dir2  -width 45 -anchor w \
            -textvariable Ventana::window(labelDir)]
    set etq_file2 [label $marco_fil.etq_file2 -width 45 -anchor w \
            -textvariable Ventana::window(labelFile)]
    set etq_que2  [label $marco_que.etq_que2  -width 30 -anchor w \
            -textvariable Ventana::Rizo::curlReport(left)]

    set progressBar [Progreso::ProgressBar $marco_bar.progress]

    set window(top)  $win
    set window(url)  $etq_url2
    set window(dir)  $etq_dir2
    set window(file) $etq_file2
    set window(left) $etq_que2
    set window(bar)  $progressBar

    set pause    [button $botones.pause    -textvariable labelButtons(pause)\
        -command ::Ventana::PauseDownloading  -width 8]
    set cancelar [button $botones.cancelar -textvariable labelButtons(skip) \
        -command "::Ventana::CancelDownloading skip" -width 8]
    set window(cancel) $cancelar

    pack $marco     -ipady 5
    pack $botones   -fill x -padx 6
    pack $marco_in  -padx 10 -side bottom
    pack $marco_url $marco_in $marco_dir $marco_fil $marco_que $marco_bar \
            -fill x
    pack $etq_url   $etq_url2  -side left -pady 2 -padx 2
    pack $etq_dir   $etq_dir2  -side left -pady 2 -padx 2
    pack $etq_file  $etq_file2 -side left -pady 2 -padx 2
    pack $etq_que   $etq_que2  -side left -pady 2 -padx 2
    pack $progressBar

    pack $cancelar $pause -side right -padx 3 -pady 5

    update
    wm protocol $win WM_DELETE_WINDOW ::Ventana::CancelDownloading

    return
}

###############################################################################
# ShowWindow
#   If needed, maps the downloading window, and sets its title bar
#
# Parameter
#   which: either 'head', 'down' or 'process', depending on the action taking
#          place
###############################################################################
proc ShowWindow {which} {
    global labelTitles
    variable window

    if {$::getleftState(downloading)==0} {
        if {$::DEBUG==1} {
            tk_messageBox -title Error -icon error -type ok \
                    -message "Pringamos en ShowWindow: $which"
        }
        return
    }   

    if {![winfo exists .bajando]} {
        DownloadWindow
    } else {
        set state [wm state .bajando]
        if {$state=="withdrawn"} {
            wm deiconify .bajando
        }
    }
    switch -exact $which {
        head {
            wm title .bajando $labelTitles(connect)
            $window(cancel) configure -state normal
        }
        down {
            wm title .bajando $labelTitles(download)
            $window(cancel) configure -state normal
        }
        process {
            wm title .bajando $labelTitles(process)
            $window(cancel) configure -state disabled
        }
    }
    return
}

###############################################################################
# FileDownloading
#   Creates a nice window to show info about how the downloading goes
#
# Parameters:
#    fichero: file in which the link will be saved.
#    enlace:  file url.
#    madre:   referer page for the link
###############################################################################
proc FileDownloading {fichero enlace madre} {
    global getleftState directories
    global labelTitles
    global Ventana::Rizo::meta
    global Ventana::Rizo::curlReport
    variable window
    variable file

    if {$getleftState(downloading)==0} {
        return
    }

    set file   $fichero

    ShowWindow down
    FileStrings $enlace

    set curlReport(end)   0
    set curlReport(long)  1
    set curlReport(pause) 0

    if {([file exists $fichero])&&($meta(versionServer)>=1.1)} {
        set size [file size $fichero]
        if {($size>$meta(totalBytes))&&($meta(totalBytes)!=0)} {
            file delete $fichero
        } elseif {($size==$meta(totalBytes))&&($meta(totalBytes)!=0)} {
            after 1000 "set ::Ventana::Rizo::curlReport(nextFile) 1"
            return
        }
        Ventana::Rizo::ResumeRequest $fichero $enlace $madre
    } else {
        Ventana::Rizo::DataRequest $fichero $enlace $madre
    }
    BarUpdate

    return
}

###############################################################################
# HeadDownloading
#    Gets the headers of the url from the server.
#
# Parameters:
#    enlace: url to download
#    madre:  referer page of the link
###############################################################################
proc HeadDownloading {enlace madre} {
    global siteUrl
    variable link
    variable mother

    set link   $enlace
    set mother $madre

    ShowWindow head
    FileStrings $enlace

    if [regexp (ftp://) $enlace] {
        return
    }

    set Ventana::Rizo::curlReport(end) 0
    Ventana::Rizo::HeadRequest $enlace $madre
    tkwait variable Ventana::Rizo::curlReport(end)

    return
}

###############################################################################
# PauseDownloading
#    Takes control when you click on the 'Pause' button
###############################################################################
proc PauseDownloading {} {
    global labelButtons
    variable window
    variable file
    variable link
    variable mother

    set Ventana::Rizo::curlReport(pause) 1
    set boton ".bajando.botones.pause"
    $boton configure -textvariable labelButtons(resume) -command "
        set Ventana::Rizo::curlReport(pause) 0
        .bajando.botones.pause configure -textvariable labelButtons(pause) \
                -command ::Ventana::PauseDownloading
        Progreso::ProgressBarReset $window(bar)
        Ventana::Rizo::ResumeRequest \"$Ventana::file\" $Ventana::link \"Ventana::$mother\"
        Ventana::BarUpdate
    "
    return
}

###############################################################################
# BarUpdate
#    Updates the progress bar.
###############################################################################
proc BarUpdate {} {
    global Ventana::Rizo::curlReport
    global Ventana::Rizo::curlError
    global Ventana::Rizo::meta
    global labelDialogs options
    variable file
    variable link
    variable mother
    variable window
    variable afterId

    if [string match $curlReport(percentage) ""] return
    Progreso::ProgressBarUpdate $window(bar) $curlReport(percentage)

    if {$curlReport(pause)==0} {
        if {($curlReport(end)==0) && ($curlReport(stop)!=1)} {
            set afterId [after 1000 Ventana::BarUpdate]
        } else {

if {$::DEBUG==1} {
    if {($curlError!=0)&&($curlError!="")} {
        tk_messageBox -icon error -title Error -type ok \
                -message "En Ventana: $curlError"
    }
}
            if {$curlError==18} {
                catch {error "Fichero Incompleto" GETLEFT \
                        "Getleft \"$labelDialogs(fileIncom)\" 001"}
                if {($options(resume)==1)&&($curlReport(stop)!=1)} {
                    Progreso::ProgressBarReset $window(bar)
                    FileDownloading $Ventana::file $Ventana::link $Ventana::mother
                    return
                }
            }
            Progreso::ProgressBarFull $window(bar)
            update
            after 75
            Progreso::ProgressBarReset $window(bar)
            set ::Ventana::Rizo::curlReport(left) ""
            $window(dir)  configure -text ""
            $window(file) configure -text ""
            incr curlReport(nextFile)
        }
    }
    return
}

}
